/** @file vdif2vdif convert VDIF data to VDIF data by changing Framesize of VDIF header 
 * @Version $Rev: 256 $
 * @Author M.Sekido
 * @Date  $Date: 2017-10-25 05:52:20 +0000 (Wed, 25 Oct 2017) $
 * $Id: vdif2vdif.cpp 256 2017-10-25 05:52:20Z vlbi $ 
 */

#include <iostream>
#include <string>
#include <vector>
using namespace std;

#include <cstdio>
#include <cstdlib>
#include <cstring>
#include "time_lib.hpp"

#include "Raw_Proto.hpp"
#include "VDIF_Header.hpp"
#include "VDIF_Frame.hpp"
#include "StaticVals.hpp"
#include "LogSet.hpp"

extern "C"{
#include <unistd.h>
#include <sys/time.h>
}

/**! "main_cmn.hpp": Common Declraration of static parameters. 
 * Should be included in main file only. */
#include "main_cmn.hpp"

const char* version_str = "$Rev: 256 $ updated $Date: 2017-10-25 05:52:20 +0000 (Wed, 25 Oct 2017) $ ";
/** General Configulation data */
Config_set vdif_conf;

//void LogSetup(const int verbose_level);

void usage(){
  cerr << "Usage: " << me << " -f <filename> [-o <outdir>] [-n <nframe/sec>]\n";
  cerr << "   [-v <level>] \n";
  cerr << "-f : input VDIF file.\n";
  cerr << "-n : Number of frames in one second.\n";
  cerr << "-v : Verbose level -1,..,4.\n";
  cerr << "-o : output directory of VDIF file. \n";
  //  cerr << "   File name including start time Tag is automatically attached.\n";
  cerr << "   Version: " << version_str << endl;
  //  cerr << "-z : Size of data reading at once.(Default= 3 MByte.)\n";
  exit(1);
}

int main(int argc, char *argv[]){
  SetMyname(argv[0]);
  //  me = (me = strrchr(argv[0], '/')) == NULL ? argv[0] : me+1; /* My name */

  string infname;
  string outfile;
  string hdstr;
  
  char logstr[256];
  Time_dat t0_cnv; // Conversion Start time
  int nf_sec = 100; // Number of frames in a second
  int verbose_level=0;
  bool log_l = false;

  int k=0;
  while (k >= 0) { 
    switch (k = getopt(argc, argv, "hf:o:n:v:")) { 
    case 'h':  usage(); break;
    case 'f':    infname = optarg;      break;
    case 'n':    nf_sec = strtol(optarg,0,10);      break;
    case 'v':    verbose_level = strtol(optarg,0,10);      
      	sprintf(logstr,"%s_v%d",me,verbose_level); 
	log_l=true;	break;
    case 'o':    outfile = optarg;     break;
    case '?' : /* Error:  character not in list */ 
      fprintf(stderr, /* Yes */ 
	      "%s WARNING:  Unknown option on command line \n", me); 
      usage();
      break; 
    case ':' : /* Error:  missing parameter */ 
      fprintf(stderr, "%s WARNING:  Missing number on command line \n", me); 
      usage();
      break; 
    }
  }

  if(infname.empty() || outfile.empty()) usage();
  //  LogSetup(verbose_level);
  LogSet logset(logstr,-1);
  if (log_l){ logset.Set_level(verbose_level); }
  logset.Start();

  
  int wfd =0;
  struct timeval tv0, tv1;  
  double total_wsize=0;
  int total_wframe=0;
  int rd_nframe=10;

  try{
    gettimeofday(&tv0,0); // Processing start time
    
    VDIF_Frame  vdif;
    if( ! vdif.ROpen(infname)){
      string msg =  "Error in opening file: " +  infname;
      throw msg;
    }
  
    vector <VDIF_Hdr> hdrarr(rd_nframe);
    if( ! (vdif.find_hdr() && vdif.Find_FirstFrame()) ){
      fprintf(stderr,"Error: Failed to find first frame.\n");
      exit(1);
    };

    //    long nframe = vdif.Read_every_hdr(// hdrarr,rd_nframe);
    // fprintf(stdout,"Total data length: %lu sec \n", hdrarr.size()); 
  
    // if(nframe < 1){
    //   string msg = "E Failed to find VDIF Header.";
    //   throw msg;
    // }
    // int iframe0 =-1;
    // for(int i=0; i< (int)hdrarr.size(); i++){
    //   if( hdrarr[i].iFrame() ==0){ iframe0 = i; break;}
    // }
    // fprintf(stdout, "Initail(iframe=0) data(%d-th frm): %s\n", 
    // 	    iframe0, hdrarr[iframe0].Str_hd(VDIF_Hdr::ONELINE).c_str());

    // if(iframe0 <0){
    //   char buf [1048];
    //   sprintf(buf,"E Failed to find Intial frame of a second within %d frames.",rd_nframe);
    //   string msg = buf;
    //   throw msg;
    // }else{      
    //   vdif.Rewind();      
    //   vdif.SeekFrames(iframe0);
    // }
  
    size_t framesize = vdif.Framesize();
    size_t hdrsize = vdif.Headsize();
    long datsize = framesize - hdrsize;
    int nch = hdrarr[0].Nch();
    int qbit = hdrarr[0].Qbit();
  
    VDIF_Frame new_vdif = vdif;
    
    new_vdif.Set_Nframe_sec(nf_sec);
    size_t new_framesize = new_vdif.Framesize();
    
    cerr << "========= Conv VDIF2VDIF ver. 1.$Rev: 256 $ start ================\n";
    double srate = (double) datsize *vdif.Nframe_sec() * 8/qbit/nch*1.e-6;
    fprintf(stdout," Framesize:%lu Byte (%dch/%dbit) \n", framesize,nch,qbit);
    // fprintf(stdout," Nsec =%ld sec, total %lu frames, SRate(Estimate):%6.1lf Msps\n", 
    // 	    hdrarr.size(), (unsigned long)nframe,  srate); 
    fprintf(stdout," SRate(Estimate):%6.1lf Msps\n",  srate); 
    fprintf(stdout,"=== Converting frame size from %8u Byte (%8u frames/sec) ", 
	    (unsigned int)framesize, (unsigned int)vdif.Nframe_sec());
    fprintf(stdout,"to %8u Byte (%8u frames/sec) \n",
	    (unsigned int)new_framesize, (unsigned int)nf_sec); 

    RAW_Proto file_prot;
    file_prot.Set_VlbiFrame(&new_vdif);
    //    DAS_Info das((uint32_t)(srate*1.e6),nch,qbit);
    //    file_prot.Set_DasInf(das);
    file_prot.Set_Host(outfile.c_str());
    if(file_prot.Tx_Ready()){
      printf(" Converting from (%s) to (%s) \n", 
	     infname.c_str(), outfile.c_str());
    }else{
      char buf[2048];
      string msg = "E Failed to open outfile: " + outfile + ":";
      msg += strerror_r(errno,buf,2048);
      throw msg;
    }
    
    //    if(Readsize < framesize) Readsize = framesize;
    vector <uint8_t> rawdat;
    int rframe = vdif.Nframe_sec();
    int cnt=0;
    ReturnFrameInfo retinfo[2];
    while( ! vdif.EndofFile_Yes() ){
      retinfo[0] = vdif.Read_frame(hdrarr,rawdat,rframe);
      // if( retinfo[0].State() == ReturnFrameInfo::NG){
      // 	string msg = "E Failed to read datafile";
      // 	throw msg;
      // 	break;
      // }      
      switch(retinfo[0].State()){
      case ReturnFrameInfo::NG:
	{
	  string msg1 = "E Read_frame():" + retinfo[0].msg;
	  throw msg1;
	}
	break;
      case ReturnFrameInfo::END:
	{
	  string msg1 = "End of File.";
	  throw msg1;
	}
	break;
      case ReturnFrameInfo::JUMP:
	{
	  int64_t jump_offst = retinfo[0].frame_intvl * datsize;
	  lseek(wfd, jump_offst,SEEK_CUR);
	}
      case ReturnFrameInfo::OK:
	{
	  new_vdif.Set_Epoch(hdrarr[0].Epoch().Time_t());
	  size_t wsize = file_prot.TX1sec(hdrarr[0].Epoch().Time_t(), rawdat);
	  if(wsize == 0){
	    string msg= "E failed to write.";
	    throw msg;
	  }
	  total_wsize += (double)wsize;
	  total_wframe+= new_vdif.Nframe_sec();
	}	
	break;
      } // End Switch 
      retinfo[1] = retinfo[0];
      if(cnt % 10==0)
	fprintf(stderr,"wrote  %.0lf MBytes. \r", (total_wsize*1.e-6)); 
      cnt++;
    } // End While

    
  }// End of Try

  
  catch(string msg){
    cerr << msg << endl;
  }

  gettimeofday(&tv1,0); // Processing End time
  double dt = (tv1.tv_sec - tv0.tv_sec)+  (tv1.tv_usec - tv0.tv_usec)*1.e-6 ;
  double wrate = total_wsize*1.e-6/dt; // MB/sec
  fprintf(stderr,"wrote  %.1lf MBytes. %d frames\n", total_wsize*1.e-6, total_wframe); 
  fprintf(stderr," %.3lf sec elapsed(%.1lf MB/s). Finished.\n", dt, wrate);
  cerr << "==========================================\n";
  
}



void LogSetup(const int verbose=-1){

  string logdir = "/tmp/";
  //  string error_file = logdir + me +"_error.txt";
   
  string debug_file = logdir + me + "_dbg.txt";
  Logsystem::debug_logger()->set_stream(debug_file);
  string info_file = logdir + me + "_info.txt";
  Logsystem::info_logger()->set_stream(info_file);
  string warn_file = logdir + me + "_warn.txt";
  Logsystem::warning_logger()->set_stream(warn_file);
  string error_file = logdir + me + "_err.txt";
  Logsystem::error_logger()->set_stream(error_file);

  uid_t vdif_uid(0);
  gid_t vdif_gid(0);
  //  bool chown_flag = false;
  
  if(vdif_conf.Check_Key("Owner") && vdif_conf.Check_Key("Group")){
    vdif_uid = atoi(vdif_conf.Value("Owner").c_str());
    vdif_gid = atoi(vdif_conf.Value("Group").c_str());
    if(vdif_uid * vdif_gid > 0){
      //      chown_flag = true;
    }else{
      cerr << "E Vdif_Config was used but 'Owner' or 'Group' Entry is missing, thus not used" << endl; 
      //      chown_flag = false;
    }
  }
  
  int err_arr[4] = {0,0,0,0};
  vector <string> flist;
  flist.push_back(error_file);
  flist.push_back(info_file);
  flist.push_back(debug_file);
  flist.push_back(warn_file);

  switch(verbose){
  case -1:
    cerr << " No Logging." << endl;
    break;
  case 0:
    Logsystem::debug_logger()->enable_logging(verbose); 
    Logsystem::info_logger()->enable_logging(verbose);
    Logsystem::error_logger()->enable_logging( verbose );
    err_arr[0] = chown(error_file.c_str(), vdif_uid, vdif_gid);
    err_arr[1] = chown(info_file.c_str(),  vdif_uid, vdif_gid);
    err_arr[2] = chown(debug_file.c_str(), vdif_uid, vdif_gid);
    
    break;
  case 1:
    Logsystem::debug_logger()->enable_logging(verbose); 
    Logsystem::info_logger()->enable_logging(verbose);
    Logsystem::error_logger()->enable_logging(verbose);
    Logsystem::warning_logger()->enable_logging(verbose);
    err_arr[0] = chown(error_file.c_str(),   vdif_uid, vdif_gid);
    err_arr[1] = chown(info_file.c_str(),    vdif_uid, vdif_gid);
    err_arr[2] = chown(debug_file.c_str(),   vdif_uid, vdif_gid);
    err_arr[3] = chown(warn_file.c_str(), vdif_uid, vdif_gid);
    break;
  case 2:
    Logsystem::debug_logger()->enable_logging(verbose); 
    Logsystem::info_logger()->enable_logging(verbose);
    Logsystem::error_logger()->enable_logging(verbose);
    Logsystem::warning_logger()->enable_logging(verbose);
    err_arr[0] = chown(error_file.c_str(),   vdif_uid, vdif_gid);
    err_arr[1] = chown(info_file.c_str(),    vdif_uid, vdif_gid);
    err_arr[2] = chown(debug_file.c_str(),   vdif_uid, vdif_gid);
    err_arr[3] = chown(warn_file.c_str(), vdif_uid, vdif_gid);
    break;
  case 3:
    Logsystem::debug_logger()->enable_logging(verbose); 
    Logsystem::info_logger()->enable_logging(verbose);
    Logsystem::error_logger()->enable_logging(verbose);
    Logsystem::warning_logger()->enable_logging(verbose);
    err_arr[0] = chown(error_file.c_str(),   vdif_uid, vdif_gid);
    err_arr[1] = chown(info_file.c_str(),    vdif_uid, vdif_gid);
    err_arr[2] = chown(debug_file.c_str(),   vdif_uid, vdif_gid);
    err_arr[3] = chown(warn_file.c_str(), vdif_uid, vdif_gid);
    break;
  case 4:
  default:
    Logsystem::debug_logger()->enable_logging(verbose); 
    Logsystem::info_logger()->enable_logging(verbose);
    Logsystem::error_logger()->enable_logging(verbose);
    Logsystem::warning_logger()->enable_logging(verbose);
    err_arr[0] = chown(error_file.c_str(),   vdif_uid, vdif_gid);
    err_arr[1] = chown(info_file.c_str(),    vdif_uid, vdif_gid);
    err_arr[2] = chown(debug_file.c_str(),   vdif_uid, vdif_gid);
    err_arr[3] = chown(warn_file.c_str(), vdif_uid, vdif_gid);
    break;
  }
  char msg[256];
  sprintf(msg,"Logging Start with level(%d)------------- \n",verbose);
  
  Logsystem::debug_logger()->Tstamp() << msg;
  Logsystem::info_logger()->Tstamp() << msg;
  Logsystem::error_logger()->Tstamp() << msg;
  Logsystem::warning_logger()->Tstamp() << msg;
  Logsystem::debug_logger()->Flush();
  Logsystem::info_logger()->Flush();
  Logsystem::error_logger()->Flush();
  Logsystem::warning_logger()->Flush();

  for(int i=0; i<4; i++){
    if(err_arr[i] < 0) fprintf(stderr,"E Failed to change Owner [%s]\n", flist[i].c_str());
  }

}
