#include <iostream>
#include <string>
#include <vector>
using namespace std;

#include <cstdio>
#include <cstdlib>
#include <cstring>
#include "time_lib.hpp"

#include "VDIF_Header.hpp"
#include "VDIF_Frame.hpp"
#include "conf_utl.hpp"
extern "C"{
#include <unistd.h>
#include <sys/time.h>
}

/**! "main_cmn.hpp": Common Declraration of static parameters. 
 * Should be included in main file only. */
#include "main_cmn.hpp"
const char* version = "$Revision: 243 $ -Last Update: $Date: 2017-10-12 16:09:31 +0900 (木, 12 10月 2017) $";


Config_set vdif_conf;

void LogSetup(const int verbose_level);

void usage(){
  string version_str = version;
  cerr << "Usage: " << me << " -f <filename> [-o <outdir>] \n";
  cerr << "   [-v <level>] [-s <stanme>:<ID>] [-n <sec>]\n";
  cerr << "   [-y <year>]\n";
  //  cerr << "   [-n <sec>] [-z <MB>] \n";
  cerr << " Rewriting VDIF file with checking frame-header.\n";
  cerr << "-n : Number of second to be converted. (default:0 All data)\n";
  cerr << "-f : input VDIF file.\n";
  //  cerr << "-v : Verbose level -1,..,4.\n";
  cerr << "-o : output directory of VDIF file. \n";
  cerr << "-y : Specify year for  detection of VDIF header.\n";
  cerr << "-s : Station Name and ID.\n";
  cerr << "   Simply re-writeing vdif without change framesize.\n";
  cerr << "   Version: " << version_str << endl;
  //  cerr << "-z : Size of data reading at once.(Default= 3 MByte.)\n";
  exit(1);
}

int main(int argc, char *argv[]){
  SetMyname(argv[0]);
  string stdummy;
  long nsec = 0;
  string infname, outfile;
  int year=0;

  int k=0;
  while (k >= 0) { 
    switch (k = getopt(argc, argv, "hf:o:n:y:s:")) { 
    case 'h':  usage(); break;
    case 'f':    infname = optarg;      break;
    case 'y':    year = strtol(optarg,0,10); break;
    case 'n':    nsec = strtol(optarg,0,10); break;
    case 's':    stdummy = optarg;       break;
      //   case 'v':    verbose_level = strtol(optarg,0,10);      break;
    case 'o':    outfile = optarg;     break;
    case '?' : /* Error:  character not in list */ 
      fprintf(stderr, /* Yes */ 
	      "%s WARNING:  Unknown option on command line \n", me); 
      usage();
      break; 
    case ':' : /* Error:  missing parameter */ 
      fprintf(stderr, "%s WARNING:  Missing number on command line \n", me); 
      usage();
      break; 
    }
  }

  char stname[30], stid[5];
  if( ! stdummy.empty()){
    char buf[80];
    strcpy(buf,stdummy.c_str());
    char *cp = strchr( buf, ':');
    if(cp != NULL){
      *cp='\0';
      strncpy(stname,buf, 30);
      cp++;
      strncpy(stid,cp,2); stid[2]='\0';
    }
  }
  
  if(infname.empty() || outfile.empty()) usage();
  //  LogSetup(verbose_level);

  struct timeval tv0, tv1;   
  double total_wsize=0;
  int rd_nframe=10;
 
  try{
    gettimeofday(&tv0,0); // Processing start time
    
    VDIF_Frame  vdif;
    if( ! vdif.ROpen(infname)){
      string msg =  "Error in opening file: " +  infname;
      throw msg;
    }
    
    vector <VDIF_Hdr> hdrarr(rd_nframe);
    long nframes = vdif.Read_1sec_hdr( hdrarr,rd_nframe);
    if(nframes < 1){
      string msg = "E Failed to find VDIF Header.";
      throw msg;
    }
    
    long framesize = vdif.Framesize();
    long datsize = framesize - vdif.Headsize();
    int nch = vdif.Hdr().Nch();
    int qbit = vdif.Hdr().Qbit();
    long srate = vdif.Hdr().Srate();
    double drate = (double)srate/1000000*nch*qbit/8; // MB/sec
    fprintf(stdout,"Original VDIF: %s \n", infname.c_str()); 
    fprintf(stdout,"Output   VDIF: %s \n", outfile.c_str()); 
    fprintf(stdout,"DataRate:%.3lf MB/sec\n", drate);
    fprintf(stdout,"Nframe/sec [real/ideal]=%ld/%ld, framesize=%ld (%ld)Byte, Initial iframe=%ld \n",
	    nframes,   (long unsigned int)vdif.Nframe_sec(), 
	    framesize, datsize,  (long int)hdrarr[0].iFrame());
    for(int i=0; i<rd_nframe; i++){
      cout << hdrarr[i].Str_hd(VDIF_Hdr::ONELINE) << endl;
    }
    VDIF_Hdr  outvdif_hdr=hdrarr[0];    
    VDIF_EUD_NICT eud_nict = hdrarr[0].EDV();
    eud_nict.Set_stname(stname);
    uint32_t srate_ksps = hdrarr[0].Srate_ksps();
    
    vdif.Rewind(VDIF_Frame::HEADER);
    unsigned char in_dummy[framesize];
    VDIF_Hdr *phdr=(VDIF_Hdr*)in_dummy;
    FILE* ofp = fopen(outfile.c_str(), "wb");
    if(ofp == NULL){
      cerr << "E Failed to open outfile(" << outfile << ")\n";
      exit(1);
    }
    
    Time_dat t0=hdrarr[0].Epoch(), t1=t0, t2=t0;
    long isec=-1; 
    long frm_cnt=0;

    while(!vdif.EndofFile_Yes() && isec < nsec){
      if( vdif.Read_Bytes( in_dummy, framesize ) == 0 ){
	cerr << "E : Failed to read.\n";
	break;
      }
      
      if( (year >0 && phdr->HD_Match("Y", &year) )
	  || (year == 0 && phdr->HD_Match())){
	
	t2=phdr->Epoch();
	phdr->exud_NICT=eud_nict;
	phdr->Set_Srate(srate_ksps);
	phdr->Set_Stnid(stid);

	if(fwrite( (void*)in_dummy, framesize, 1, ofp) < 1){
	  fprintf(stderr,"E Failed to write frame to outfile\n");
	  exit (1);
	}
	total_wsize +=framesize;
	frm_cnt++;
	if(nsec > 0) isec= t2-t0;
	if( t2 > t1 ){ t1=t2;
	  fprintf(stderr,"\r%s",t1.Print_time(14).c_str());
	}else  if(frm_cnt % 1000000 ==0) putc('.',stderr);

      }else{ // if HD not mached 
	if( ! vdif.find_hdr() ){
	  string msg="E Faile to find Header.";
	throw msg;
	}
      }
    } // end while
    fclose(ofp);

    gettimeofday(&tv1,0); // Processing End time
    double dt = (tv1.tv_sec - tv0.tv_sec)+  (tv1.tv_usec - tv0.tv_usec)*1.e-6 ;
    double wrate = total_wsize*1.e-6/dt; // MB/sec
    fprintf(stderr,"wrote  %.1lf MBytes. %ld frames\n", 
	    total_wsize*1.e-6, frm_cnt); 
    fprintf(stderr," %.3lf sec elapsed(%.1lf MB/s). Finished.\n", dt, wrate);
    cerr << "==========================================\n";
    
  }// End of Try
  
  
  catch(string msg){
    cerr << msg << endl;
  }
  
}
