
/** @file vdif2vsi convert VDIF data to VSI data by removing the VDIF header 
 * @Version $Rev: 257 $
 * @Author M.Sekido
 * @Date  $Date: 2017-10-25 08:00:44 +0000 (Wed, 25 Oct 2017) $
 * $Id: vdif2vsi.cpp 257 2017-10-25 08:00:44Z vlbi $ 
 */

#include <iostream>
#include <string>
#include <vector>
using namespace std;

#include <cstdio>
#include <cstdlib>
#include <cstring>
#include "time_lib.hpp"

#include "VDIF_Header.hpp"
#include "VDIF_Frame.hpp"
#include "StaticVals.hpp"
//#include "DAS_info.hpp"
//#include "VsiData.hpp"
//#include "conf_utl.hpp"

extern "C"{
#include <unistd.h>
#include <sys/time.h>
}

/**! "main_cmn.hpp": Common Declraration of static parameters. 
 * Should be included in main file only. */
#include "main_cmn.hpp"

/** General Configulation data */

Config_set vdif_conf;
const char* version_str = "$Rev: 257 $ updated $Date: 2017-10-25 08:00:44 +0000 (Wed, 25 Oct 2017) $ ";

void usage(){
  cerr << "Usage: " << me << " -f <filename> [-o <outdir>] [-s <str>] [-t <yyyydddhhmmss>]\n";
  cerr << "   [-n <sec>] [-z <MB>] \n";
  cerr << "-n : Number of second to be converted. (default:0 All data)\n";
  cerr << "-f : input VDIF file.\n";
  cerr << "-t : Data conversion start time. \n";
  cerr << "    If '-t' is not specified, Intial Epoch of the data is used for the file name.\n";
  cerr << "-s : Header string of output VDIF name. <str_yyyydddhhmmss.vsi>\n";
  cerr << "-o : output directory of VSI file. \n";
  cerr << "   File name including start time Tag is automatically attached.\n";
  cerr << "-z : Size of data reading at once.(Default= 3 MByte.)\n";
  cerr << "   Version: " << version_str << endl;
  exit(1);
}

int main(int argc, char *argv[]){
  SetMyname(argv[0]);
  //  me = (me = strrchr(argv[0], '/')) == NULL ? argv[0] : me+1; /* My name */

  string infname;
  string outfile;
  string hdstr;

  //  int nsec =0;
  size_t Readsize = 3000000; // Reading size at once;
  Time_dat t0_cnv; // Conversion Start time
  bool t0_specify=false;
  
  int k=0;
  while (k >= 0) { 
    switch (k = getopt(argc, argv, "hf:o:z:s:t:")) { 
    case 'h':  usage(); break;
    case 'f':    infname = optarg;      break;
    case 's':    hdstr = optarg;      break;
    case 'o':    outfile = optarg;     break;
    case 't':    t0_cnv.Set_time(optarg); t0_specify=true;   break;
      //    case 'n':    nsec = strtol(optarg,0,10);      break;
    case 'z':    Readsize = strtol(optarg,0,10)*1000000;      break;
    case '?' : /* Error:  character not in list */ 
      fprintf(stderr, /* Yes */ 
	      "%s WARNING:  Unknown option on command line \n", me); 
      usage();
      break; 
    case ':' : /* Error:  missing parameter */ 
      fprintf(stderr, "%s WARNING:  Missing number on command line \n", me); 
      usage();
      break; 
    }
  }

  if(infname.empty() ) usage();
  if(outfile.empty()){outfile = "./";} 
  
  int wfd =0;
  struct timeval tv0, tv1;  
  double total_wsize=0;
  int total_wframe=0;

  try{
    
    gettimeofday(&tv0,0); // Processing start time
    
    VDIF_Frame  vdif;
    if( ! vdif.ROpen(infname)){
      string msg =  "Error in opening file: " +  infname;
      throw msg;
    }
  
    vector <VDIF_Hdr> hdrarr(10);
    long nframe = vdif.Read_every_hdr(hdrarr,10);
    fprintf(stdout,"Total data length: %lu sec \n", hdrarr.size()); 
  
    if(nframe < 1){
      string msg = "E Failed to find VDIF Header.";
      throw msg;
    }
    int iframe0 =-1;
    if(t0_specify){
      for(int i=0; i< (int)hdrarr.size(); i++)
	if( hdrarr[i].iFrame() ==0 && (hdrarr[i].Epoch() >= t0_cnv) ){ 
	  iframe0 = i; break;}
    }else{
      for(int i=0; i< (int)hdrarr.size(); i++){
	if( hdrarr[i].iFrame() ==0){ iframe0 = i; break;}
      }
    }
    fprintf(stdout, "Initail data(%d frm): %s\n", 
	    iframe0, hdrarr[iframe0].Str_hd(VDIF_Hdr::ONELINE).c_str());

    if(iframe0 <0){
      string msg = "E Failed to find Intial frame of a second.";
      throw msg;
    }else{      
      vdif.Rewind();      
      vdif.SeekFrames(iframe0);
      if(outfile.at(outfile.length()-1) != '/') outfile += "/";
      char buf[1024];
      if(hdstr.empty()){hdstr = infname.substr(0,1);}
      const Time_dat &t1 = hdrarr[iframe0].Epoch();
      sprintf(buf,"%s_%s.vsi",
	      hdstr.c_str(), t1.Print_time(13).c_str()  );
      outfile += buf; 
    }
  
    size_t framesize = vdif.Framesize();
    size_t hdrsize = vdif.Headsize();
    long datsize = framesize - hdrsize;
    int nch = hdrarr[0].Nch();
    int qbit = hdrarr[0].Qbit();
    //    int npt_frame = datsize * 8/qbit/nch; // #point per frame;
    //    const int Ncmp = 100000;
    //    const int Nframe_comp = (npt_frame > Ncmp) ? Ncmp : npt_frame;
    //    int lbound(0), ndiv(2), istep(0);
  
    cerr << "========= Conv VDIF2VDI ver. 1.$Rev: 257 $ start ================\n";
    double srate = (double) datsize *vdif.Nframe_sec() * 8/qbit/nch*1.e-6;
    fprintf(stdout," Framesize:%lu Byte (%dch/%dbit) \n", framesize,nch,qbit);
    fprintf(stdout," Nsec =%ld sec, total %lu frames, SRate(Estimate):%6.1lf Msps\n", 
	    hdrarr.size(), (unsigned long)nframe,  srate); 

    uint wmode= S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP | S_IROTH; 
    uint wflag = O_CREAT  | O_LARGEFILE | O_WRONLY | O_TRUNC;  
    wfd = open(outfile.c_str(), wflag, wmode);
    if ( wfd <= 0 ){
      char buf[2048];
      string msg = "E Failed to open outfile: " + outfile + ":";
      msg += strerror_r(errno,buf,2048);
      throw msg;
    }
    else  {
      printf(" Converting from (%s) to (%s) \n", 
	     infname.c_str(), outfile.c_str());
    }
    if(Readsize < framesize) Readsize = framesize;
    vector <uint8_t> rawdat;
    int rframe = Readsize / framesize;
    int cnt=0;
    ReturnFrameInfo retinfo[2];
    while( ! vdif.EndofFile_Yes() ){
      retinfo[0] = vdif.Read_frame(hdrarr,rawdat,rframe);
      if( retinfo[0].State() == ReturnFrameInfo::NG){
	string msg = "E Failed to read datafile";
	throw msg;
	break;
      }      
      switch(retinfo[1].State()){
      case ReturnFrameInfo::NG:
	{
	  string msg = "E Read_frame():Impossible NG";
	  throw msg;
	}
	break;
      case ReturnFrameInfo::END:
	{
	  string msg = "EDN of FILE.";
	  throw msg;
	}
	break;
      case ReturnFrameInfo::JUMP:
	{
	  int64_t jump_offst = retinfo[1].frame_intvl * datsize;
	  lseek(wfd, jump_offst,SEEK_CUR);
	}
      case ReturnFrameInfo::OK:
	{
	  size_t wsize = retinfo[0].nframe * datsize;
	  int  wrsize = write(wfd,&rawdat[0],wsize);
	  if(wrsize <0){
	    string msg= "E failed to write.";
	    throw msg;
	  }
	  total_wsize += (double)wrsize;
	  total_wframe+= retinfo[0].nframe;
	}	
	break;
      } // End Switch 
      retinfo[1] = retinfo[0];
      if(cnt % 10==0)
	fprintf(stderr,"wrote  %.0lf MBytes. \r", (total_wsize*1.e-6)); 
      cnt++;
    } // End While

    
  }// End of Try

  
  catch(string msg){
    cerr << msg << endl;
  }

  if(wfd > 0) close(wfd);
    gettimeofday(&tv1,0); // Processing End time
    double dt = (tv1.tv_sec - tv0.tv_sec)+  (tv1.tv_usec - tv0.tv_usec)*1.e-6 ;
    double wrate = total_wsize*1.e-6/dt; // MB/sec
    fprintf(stderr,"wrote  %.1lf MBytes. %d frames\n", total_wsize*1.e-6, total_wframe); 
    fprintf(stderr," %.3lf sec elapsed(%.1lf MB/s). Finished.\n", dt, wrate);
    cerr << "==========================================\n";
    
}


