
/** @file vdif_chk is test reading VDIF header data 
 * @Version $Rev: 246 $
 * @Author M.Sekido
 * @Date  $Date: 2017-10-17 07:28:22 +0900 (火, 17 10月 2017) $
 * $Id: vdif_mrg.cpp 246 2017-10-16 22:28:22Z vlbi $ 
 */

const char* version = "$Revision: 246 $ -Latest Update: $Date: 2017-10-17 07:28:22 +0900 (火, 17 10月 2017) $";

#include <iostream>
#include <string>
#include <vector>
using namespace std;

#include <cstdio>
#include <cstdlib>
#include <boost/algorithm/string.hpp>

#include "StaticVals.hpp"
#include "VDIF_Header.hpp"
#include "VDIF_Frame.hpp"
#include "DAS_info.hpp"
// #include "VsiData.hpp"
// #include "hist_g.hpp"
#include "conf_utl.hpp"
#include "main_cmn.hpp"

extern "C"{
#include <unistd.h>
}
void usage(){
  fprintf(stderr,"Usage: %s -f <vdif fike>:<vdif fike>[:<vdif fike>:<vdif fike>[:<vdif fike>:<vdif fike>:<vdif fike>:<vdif fike>]] \n",me);
  fprintf(stderr,"          -o <outfile> [-s <stname]>] [-i <stid>] \n");
  fprintf(stderr," Merging 8 vdif file to single thread vdif file. \n");
  fprintf(stderr," 1-8ch of file1-4 and 9-16ch of file6-8 are marged into one file with 64 channels. \n");
  fprintf(stderr,"   First or last half set of channels of 8 files are marged into one file.\n");
  fprintf(stderr," -s : Station Name to be recorded in extended user data. \n");
  fprintf(stderr," -i : Station ID 2 characters.\n");
  //  fprintf(stderr," -c : When Number of input files are 8, 0:First(default),1:Last half of word is copied.\n");
  fprintf(stderr," Ver. %s \n",version);
  exit(1);
}
bool Sync_frame(vector <VDIF_Frame>& vdif_arr); ///< Synchronize epoch and _fp
bool Check_Homo(vector <VDIF_Frame>& vdif_arr); ///< Check Homogenity 
bool Check_eof(vector <VDIF_Frame>& vdif_arr);
  
int main(int argc, char* argv[]){
  SetMyname(argv[0]);
  string instr, outfile;
  string stname, stid;
  int hlf_word=0; // first or last half word is copied.
  int k=0;
  while (k >= 0) { 
    switch (k = getopt(argc, argv, "hf:o:s:i:")) { 
    case 'h':  usage(); break;
    case 'f':   instr = optarg;      break;
    case 'o':   outfile = optarg;   break;
    case 'c':   hlf_word = strtol(optarg,0,10);   break;
    case 's':   stname = optarg;   break;
    case 'i':   stid = optarg;   break;
    case '?' : /* Error:  character not in list */ 
      fprintf(stderr, "%s WARNING:  Unknown option on command line \n", me); 
      usage();
      break; 
    case ':' : /* Error:  missing parameter */ 
      fprintf(stderr, "%s WARNING:  Missing number on command line \n", me); 
      usage();
      break; 
    }
  }

  if( instr.empty() || outfile.empty()) usage();

  vector <string> infiles;  
  boost::algorithm::split(infiles, instr, boost::is_any_of(":"));
  int nfiles = infiles.size();
  if( nfiles != 8  ){
    cerr << " input file must be 8.\n"; exit(1);
    usage();
  }
  const char *msg[2] = {"First", "Latter"};
  fprintf(stdout, "Converting %s half words of every files are merged into one. \n", msg[hlf_word] );
  for(int i=0; i<8; i++){ fprintf(stdout,"%d: %s\n",i+1,infiles[i].c_str());  }
  
  vector <VDIF_Frame>  in_vdif( nfiles );
  
  for(int i=0; i<8 ; i++){
    if( ! in_vdif[i].ROpen(infiles[i]) ) {
      fprintf(stderr, "E main(): Failed to open vdif file[%d]=%s\n",
	      i,infiles[i].c_str());   exit(1);
    }
    if( ! ( in_vdif[i].find_hdr() &&  
	    in_vdif[i].Nframe_sec_estimate()) ){  cerr << "E Failed to find VDIF Header.\n";
      exit(1);
    }else{
      fprintf(stderr,"vdif[%d] Hdr-Found:Offset=%d \n",i, in_vdif[i].Hdr_Offset() );
    }
    
  } // for loop
    
  /** Sync time and iframe*/
  if(! Sync_frame(in_vdif) ) {
    cerr <<  "E main:Sync_frame(): failed to sync.\n"; exit(1);
  }
  /** Check homogenious data property **/
  if( ! Check_Homo(in_vdif)) { 
    cerr <<  "E main:Check_Homo(): failed checking homogenity of files.\n"; exit(1);
  }
 
  int headsize = in_vdif[0].Headsize();;
  int datasize = in_vdif[0].Framesize() - headsize;
  int out_datasize = datasize * nfiles/2 ;
  
  const uint8_t *dataframe= new uint8_t[out_datasize + headsize];
  VDIF_Hdr *phdr = (VDIF_Hdr*)dataframe;
  VDIF_Hdr vhdr_dmy;
  vhdr_dmy.Param_Cpy(in_vdif[0].Hdr());
  const uint8_t* dataptr_out = (uint8_t*)dataframe + headsize;

  if(! stname.empty()) vhdr_dmy.exud_NICT.Set_stname(stname.c_str());
  if(! stid.empty()) vhdr_dmy.Set_Stnid(stid.c_str());
  if( in_vdif[0].Nframe_sec() == 0){
    cerr << "E main: nframe_sec,srate of input file is not clear\n";
    exit (1);
  }
  long nframe_sec_in = in_vdif[0].Nframe_sec();
  int qbit = in_vdif[0].Hdr().Qbit();
  int nch = in_vdif[0].Hdr().Nch();
  long bits_stream = qbit*nch;
  if( in_vdif[0].Hdr().Data_Type() == VDIF_Hdr::Complex_Sample )  bits_stream *= 2;

  // if(bits_stream != 8 ){
  //   fprintf(stderr,"E main: Bit(%d)xNch(%d) < 8 is not suppored in this program.",qbit,nch);
  //   exit(1);
  // }

  long nframe_sec_out=0;
  nframe_sec_out = nframe_sec_in;
  uint32_t srate = nframe_sec_in * datasize * 8/qbit/nch;
  vhdr_dmy.Set_Srate(srate/1000); // Set in ksps
  vhdr_dmy.Set_Nch( nch *nfiles/2 ); // Set in ksps
  vhdr_dmy.Set_Epoch( in_vdif[0].Hdr().Epoch() );
  long framesize_out = out_datasize + headsize;
  vhdr_dmy.Set_FrameLen( framesize_out  );

  vector <uint8_t> rawbuf_in[nfiles];
  vector <VDIF_Hdr> hdrarr_in[nfiles];
  ReturnFrameInfo retinf[nfiles];
  for(int i=0; i<8; i++){
    retinf[i] = in_vdif[i].Read_frame(hdrarr_in[i], rawbuf_in[i]);
    if( retinf[i].State() != ReturnFrameInfo::OK ){
      fprintf(stderr, "E main:in_vdif[%d].Read_frame(): Failed. \n", i);
      exit(1);
    }
  }

  struct timeval tv0, tv1;   
  gettimeofday(&tv0,0); // Processing start time
    
  int32_t iframe0= in_vdif[0].Hdr().iFrame();
  int32_t iframe= iframe0;
  Time_dat t1 = in_vdif[0].Hdr().Epoch();
  Time_dat t2=t1;
  long cpysize = bits_stream/8/2; // bytes;
  
  FILE* outfp=NULL;
  if( (outfp = fopen(outfile.c_str(),"w") ) ==NULL){
    fprintf(stderr,"E Failed to open outfile(%s)\n",outfile.c_str());
    exit(1);
  }
  //  phdr->Param_Cpy(vhdr_dmy);
  *phdr = vhdr_dmy;
  long frm_cnt=0;
  long total_wsize=0;
  int ifrm_mrgn=100;

  while(! Check_eof(in_vdif)){
    if( iframe +ifrm_mrgn < iframe0){
      t2 = t1 + 1;
      t1=t2;
      fprintf(stderr,"%s\n",t2.Print_time(16).c_str());
    }
    phdr->Set_Epoch( t2 );
    phdr->Set_iFrame( iframe );
    bzero((void*)dataptr_out, out_datasize);
    bool cpy_exec_flag = false;
    bool valid_flag=true;
    for(int ifile=0; ifile<8; ifile++){

      if(retinf[ifile].State() != ReturnFrameInfo::OK){
	if(in_vdif[ifile].EndofFile_Yes()){
	  fprintf(stderr, "I vdif[%d] End-of-File. End.\n", ifile);
	  break;
	}
	fprintf(stderr, "E main:in_vdif[%d].Read_frame(): Failed. \n", ifile);
	exit(1);
      }

      while( hdrarr_in[ifile][0].iFrame() < iframe &&
	 hdrarr_in[ifile][0].Epoch() == t2 ){ 
	retinf[ifile] = 
	  in_vdif[ifile].Read_frame(hdrarr_in[ifile], rawbuf_in[ifile]);
	valid_flag=false;
      }
      if( hdrarr_in[ifile][0].iFrame() > iframe ){ valid_flag=false; continue;}

      if( hdrarr_in[ifile][0].Epoch() > t2 ){ valid_flag=false;	continue;  }
      if( hdrarr_in[ifile][0].Epoch() < t2 ){
	fprintf(stderr,"E  hdr[%d],Epoch(%s) < t2(%s)\n",ifile,
		hdrarr_in[ifile][0].Epoch().Print_time(16).c_str(),t2.Print_time(16).c_str());
	exit(1);
      }

      long bytcnt=datasize;
      int skip_byte = cpysize*2;
      hlf_word = ifile < 4 ? 0 : 1;
      uint8_t *inptr  = (uint8_t*)&rawbuf_in[ifile][0] + hlf_word * cpysize;;
      uint8_t *outptr = (uint8_t*)dataptr_out + ifile * cpysize;
      while(bytcnt >0){
	memcpy((void*)outptr, inptr, cpysize);
	outptr += cpysize * nfiles;
	inptr += skip_byte;
	bytcnt -= skip_byte;
      }
      valid_flag = (hdrarr_in[ifile][0].Valid() && valid_flag ) ? true : false;
      retinf[ifile] = 
	in_vdif[ifile].Read_frame(hdrarr_in[ifile], rawbuf_in[ifile]);
      cpy_exec_flag = true;
    } // for loop nfiles
    
    if(cpy_exec_flag){
      phdr->Set_valid_flag(valid_flag);
      if( fwrite(dataframe, framesize_out, 1,outfp) < 1){
	cerr << "E main: Failed to write data\n";
	exit (1);
      }
      total_wsize += framesize_out;
    }
    
    iframe0=iframe;
    iframe = hdrarr_in[0][0].iFrame();
    iframe %= nframe_sec_out;
    frm_cnt++;
    
  } // Loop for write
  fclose(outfp);

  gettimeofday(&tv1,0); // Processing End time
  double dt = (tv1.tv_sec - tv0.tv_sec)+  (tv1.tv_usec - tv0.tv_usec)*1.e-6 ;
  double wrate = total_wsize*1.e-6/dt; // MB/sec
  fprintf(stderr,"wrote  %.1lf MBytes. %ld frames\n", 
	  total_wsize*1.e-6, frm_cnt); 
  fprintf(stderr," %.3lf sec elapsed(%.1lf MB/s). Finished.\n", dt, wrate);
  cerr << "==========================================\n";
  
  delete [] dataframe;
}

bool Check_eof(vector <VDIF_Frame>& vdif_arr){
  int nfiles = vdif_arr.size();
  for(int i=0; i<nfiles; i++){
    if( vdif_arr[i].EndofFile_Yes()) return true;
  }
  return false;
}

bool Check_Homo(vector <VDIF_Frame>& vdif_arr){
  int nfiles = vdif_arr.size();
  bool ret = true;
  for(int i=1; i<nfiles; i++){
    if(vdif_arr[0].Framesize() != vdif_arr[i].Framesize()) ret = false;
    if( vdif_arr[0].Hdr().Nch() != vdif_arr[i].Hdr().Nch() ) ret = false;
    if( vdif_arr[0].Hdr().Qbit() != vdif_arr[i].Hdr().Qbit() ) ret = false;
  }
  return ret;
}

bool Sync_frame(vector <VDIF_Frame>& vdif_arr){
  int nfiles = vdif_arr.size();
  int nsync=0;
  uint32_t sec_max(0);
  int iframe_max(0);
  while(nsync != nfiles*2 ){
    if(sec_max != 0) nsync=2;
    for(int i=0; i<nfiles; i++){
      if( sec_max < vdif_arr[i].Hdr().Sec_Ref()){
	sec_max = vdif_arr[i].Hdr().Sec_Ref();
	iframe_max =i;
	nsync=2;
	continue;
      }

      while( sec_max > vdif_arr[i].Hdr().Sec_Ref()){
	if ( ! ( vdif_arr[i].SeekFrames(1) &&
		 vdif_arr[i].Find_FirstFrame()) ){
	  fprintf(stderr, "E main:Sync_frame(): failed at vdif[%d]\n",i);
	  return false;
	}
      } // while loop
      if( sec_max == vdif_arr[i].Hdr().Sec_Ref() ){
	nsync++;
	int dframe = vdif_arr[iframe_max].Hdr().iFrame() - vdif_arr[i].Hdr().iFrame();
	while(dframe > 0){
	  if( !( vdif_arr[i].SeekFrames(dframe) && vdif_arr[i].find_hdr() )){
	    fprintf(stderr, "E main:Sync_frame(): failed at vdif[%d]\n",i);
	    return false;
	  }
	  dframe = vdif_arr[iframe_max].Hdr().iFrame() - vdif_arr[i].Hdr().iFrame();
	}
	if(dframe ==0)	nsync++;
	else iframe_max = i;
	
      } // if
    } // for loop
  } // while loop

  bool ret=true;
  for(int i=0;i<nfiles; i++){
    if( vdif_arr[i].Hdr().Sec_Ref() !=  vdif_arr[iframe_max].Hdr().Sec_Ref() ) ret=false;
    if( vdif_arr[i].Hdr().iFrame()  !=  vdif_arr[iframe_max].Hdr().iFrame() ) ret=false;
  }

  return ret;
}


